import logging

import click

# Need this so we can eval() below
from .regions import *  # noqa
from .specs import *  # noqa


@click.group(invoke_without_command=True)
@click.option(
    "--log-level",
    default="INFO",
    type=click.Choice(
        ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"], case_sensitive=False
    ),
)
@click.version_option()
@click.pass_context
def cli(ctx, log_level: str):
    """ScanSpec library command line interface."""

    level = getattr(logging, log_level.upper(), None)
    logging.basicConfig(format="%(levelname)s:%(message)s", level=level)

    # if no command is supplied, print the help message
    if ctx.invoked_subcommand is None:
        click.echo(cli.get_help(ctx))


@cli.command()
@click.argument("spec")
def plot(spec: str):
    """Plot a ScanSpec"""
    from scanspec.plot import plot_spec

    for letter in "abcuvwxyz":
        locals()[letter] = letter
    eval_spec = eval(spec)
    plot_spec(eval_spec)



from inspect import Parameter, signature
from typing import Iterator, List, Optional, Tuple

from matplotlib.sphinxext.plot_directive import PlotDirective
from pydantic import BaseModel
from pydantic.decorator import ValidatedFunction
from pydantic.typing import display_as_type

from scanspec.specs import _modifier


def validated_function_ignores(vd: ValidatedFunction) -> Iterator[str]:
    # Pydantic adds args and kwargs, remove them if not in original func
    parameters = signature(vd.raw_function).parameters
    kinds = {p.kind for p in parameters.values()}
    if Parameter.VAR_POSITIONAL not in kinds:
        yield vd.v_args_name
    if Parameter.VAR_KEYWORD not in kinds:
        yield vd.v_kwargs_name
    # Ignore the first arg as it is the classname
    # This wouldn't work for @staticmethod, but we don't have any
    yield list(parameters)[0]


def get_model_ignores(what: str, obj) -> Tuple[Optional[BaseModel], List[str]]:
    model = None
    ignores = []
    if what == "class" and issubclass(obj, BaseModel):
        # If we are a BaseModel, use this
        model = obj
    elif what == "method" and issubclass(getattr(obj, "model", type), BaseModel):
        # If we have been decorated with validate_arguments, use the model
        model = obj.model
        ignores = list(validated_function_ignores(obj.vd))
    if model:
        for name, field in model.__fields__.items():
            if field.field_info.const and not field.required:
                # Suppress this one as it is probably used like this:
                # https://pydantic-docs.helpmanual.io/usage/types/#literal-type
                ignores.append(name)
    return model, ignores


def process_docstring(app, what, name, obj, options, lines):
    model, ignores = get_model_ignores(what, obj)
    if model:
        # Insert it after the first gap
        try:
            index = lines.index("") + 1
        except ValueError:
            lines.append("")
            index = len(lines)
        # Add types from each field
        for name, field in model.__fields__.items():
            if name not in ignores:
                # UnionModifier expanded this so creation from
                # JSON works, put the original back for docs
                if field.type_ is _modifier.spec_union:
                    type_string = "Spec"
                elif field.type_ is _modifier.region_union:
                    type_string = "Region"
                else:
                    type_string = display_as_type(field.type_)
                description = field.field_info.description
                if type_string == "ConstrainedIntValue":
                    type_string = "int"
                    translate = dict(
                        gt=">", ge=">=", lt="<", le="<=", multiple_of="multiple of"
                    )
                    conditions = []
                    for attr, text in translate.items():
                        v = getattr(field.type_, attr)
                        if v:
                            conditions.append(f"{text} {v}")
                    if conditions:
                        description += f". Must be {' ,'.join(conditions)}"
                lines.insert(
                    index, f":param {type_string} {name}: {description}",
                )
                index += 1
        lines.insert(index, "")


def process_signature(app, what, name, obj, options, signature, return_annotation):
    model, ignores = get_model_ignores(what, obj)
    if model:
        # Recreate signature from the model
        args = []
        for name, field in model.__fields__.items():
            if name not in ignores:
                if field.required:
                    args.append(name)
                else:
                    args.append(f"{name}={field.default!r}")
                # TODO: what about *args and **kwargs in functions?
        signature = f'({", ".join(args)})'
    return signature, return_annotation


class ExampleSpecDirective(PlotDirective):
    def run(self):
        self.content = (
            ["# Example Spec", "", "from scanspec.plot import plot_spec"]
            + [str(x) for x in self.content]
            + ["plot_spec(spec)"]
        )
        return super().run()


def setup(app):
    app.add_directive("example_spec", ExampleSpecDirective)
    app.connect("autodoc-process-signature", process_signature)
    app.connect("autodoc-process-docstring", process_docstring)

    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from typing import Any, Callable, Dict, Iterator, List, Type, TypeVar

import numpy as np
from pydantic import BaseModel
from pydantic.fields import Field, FieldInfo


# These are used in the definition of the Schema
# It allows the class to be inferred from the serialized "type" field
class _WithTypeMetaClass(type(BaseModel)):  # type: ignore
    def __new__(mcs, name, bases, namespace, **kwargs):
        # Override type in namespace to be the literal value of the class name
        namespace["type"] = Field(name, const=True)
        return super().__new__(mcs, name, bases, namespace, **kwargs)


class _WithType(BaseModel, metaclass=_WithTypeMetaClass):
    """BaseModel that adds a type parameter from class name."""

    type: str

    class Config:
        # Forbid any extra input
        extra = "forbid"

    def __init__(self, *args, **kwargs):
        # Allow positional args, but don't include type
        keys = [x for x in self.__fields__ if x != "type"]
        kwargs.update(zip(keys, args))
        # Work around the fact that arg=Field(default, ...) doesn't
        # work with validate_arguments
        for k, v in kwargs.items():
            if isinstance(v, FieldInfo):
                kwargs[k] = v.default
        super().__init__(**kwargs)


#: Positions map {key: positions_ndarray}
#: E.g. {xmotor: array([0, 1, 2]), ymotor: array([2, 2, 2])}
Positions = Dict[Any, np.ndarray]

#: The type of class the function will return
T = TypeVar("T")


def if_instance_do(x, cls: Type[T], func: Callable[[T], Any]):
    """If x is of type cls then return func(x), otherwise return NotImplemented.
    Used as a helper when implementing operator overloading"""
    if isinstance(x, cls):
        return func(x)
    else:
        return NotImplemented


class Dimension:
    """Represents a linear stack of positions and bounds. A list of Dimensions
    is interpreted as nested from slowest moving to fastest moving, so each
    faster Dimension will iterate once per position of the slower Dimension.
    When fly-scanning they key will traverse lower-position-upper on the fastest
    Dimension for each point in the scan.

    Args:
        positions: The centre positions of the scan for each key
        lower: Lower bounds if different from positions
        upper: Upper bounds if different from positions
        snake: If True then every other iteration of this Dimension within a
            slower moving Dimension will be reversed

    See Also:
        `what-are-dimensions`
    """

    def __init__(
        self,
        positions: Positions,
        lower: Positions = None,
        upper: Positions = None,
        snake: bool = False,
    ):
        #: The centre positions of the scan for each key
        self.positions = positions
        #: The lower bounds of each scan point for each key for fly-scanning
        self.lower = lower or positions
        #: The upper bounds of each scan point for each key for fly-scanning
        self.upper = upper or positions
        #: Whether every other iteration of this Dimension within a slower
        #: moving Dimension will be reversed
        self.snake = snake
        # Check all keys and ordering are the same
        assert list(self.positions) == list(self.lower) == list(self.upper), (
            f"Mismatching keys "
            f"{list(self.positions)} != {list(self.lower)} != {list(self.upper)}"
        )
        # Check all lengths are the same
        lengths = set(
            len(arr)
            for d in (self.positions, self.lower, self.upper)
            for arr in d.values()
        )
        assert len(lengths) <= 1, f"Mismatching lengths {list(lengths)}"

    def keys(self) -> List:
        """The keys that are present in `positions`, `lower` and `upper`
        which will move during the scan"""
        return list(self.positions.keys())

    def __len__(self) -> int:
        """The number of `positions` in the scan"""
        # All positions arrays are same length, pick the first one
        return len(list(self.positions.values())[0])

    def _dim_with(self, func: Callable[[str, Any], np.ndarray]) -> "Dimension":
        def apply_func(a: str):
            return {k: func(a, k) for k in getattr(self, a)}

        # Apply to every array in positions
        kwargs = dict(positions=apply_func("positions"), snake=self.snake)
        # If lower and upper are different, apply to those too
        if self.lower is not self.positions:
            kwargs["lower"] = apply_func("lower")
        if self.upper is not self.positions:
            kwargs["upper"] = apply_func("upper")
        return Dimension(**kwargs)

    def tile(self, reps: int) -> "Dimension":
        """Return a new Dimension that iterates self reps times

        >>> dim = Dimension({"x": np.array([1, 2, 3])})
        >>> dim.tile(reps=2).positions
        {'x': array([1, 2, 3, 1, 2, 3])}
        """
        return self._dim_with(lambda a, k: np.tile(getattr(self, a)[k], reps))

    def repeat(self, reps: int) -> "Dimension":
        """Return a new Dimension that repeats each point in self reps times

        >>> dim = Dimension({"x": np.array([1, 2, 3])})
        >>> dim.repeat(reps=2).positions
        {'x': array([1, 1, 2, 2, 3, 3])}
        """
        return self._dim_with(lambda a, k: np.repeat(getattr(self, a)[k], reps))

    def mask(self, mask: np.ndarray) -> "Dimension":
        """Return a new Dimension that produces only points from self in the
        mask

        >>> dim = Dimension({"x": np.array([1, 2, 3])})
        >>> dim.mask(np.array([1, 0, 1])).positions
        {'x': array([1, 3])}
        """
        indices = mask.nonzero()[0]
        return self._dim_with(lambda a, k: getattr(self, a)[k][indices])

    def copy(self) -> "Dimension":
        """Return a shallow copy of the current Dimension (dicts copied,
        arrays within them are not)"""
        return self._dim_with(lambda a, k: getattr(self, a)[k])

    def _check_dim(self, other: "Dimension"):
        assert isinstance(other, Dimension), f"Expected Dimension, got {other}"
        assert self.snake == other.snake, "Snake settings don't match"

    def concat(self, other: "Dimension") -> "Dimension":
        """Return a new Dimension with arrays from self and other concatenated
        together. Require both Dimensions to have the same keys and snake
        settings

        >>> dim = Dimension({"x": np.array([1, 2, 3])})
        >>> dim2 = Dimension({"x": np.array([5, 6, 7])})
        >>> dim.concat(dim2).positions
        {'x': array([1, 2, 3, 5, 6, 7])}
        """
        self._check_dim(other)
        assert self.keys() == other.keys(), f"Keys {self.keys()} != {other.keys()}"
        return self._dim_with(
            lambda a, k: np.concatenate((getattr(self, a)[k], getattr(other, a)[k]))
        )

    def zip(self, other: "Dimension") -> "Dimension":
        """Return a new Dimension with arrays from keys of self and other
        merged together. Require both Dimensions to not share keys, and
        to have the snake settings

        >>> dimx = Dimension({"x": np.array([1, 2, 3])})
        >>> dimy = Dimension({"y": np.array([5, 6, 7])})
        >>> dimx.zip(dimy).positions
        {'x': array([1, 2, 3]), 'y': array([5, 6, 7])}
        """
        self._check_dim(other)
        overlapping = list(set(self.keys()).intersection(other.keys()))
        assert not overlapping, f"Zipping would overwrite keys {overlapping}"
        # rely on the constructor to check lengths
        dim = Dimension(
            positions={**self.positions, **other.positions},
            lower={**self.lower, **other.lower},
            upper={**self.upper, **other.upper},
            snake=self.snake,
        )
        return dim


def squash_dimensions(
    dimensions: List[Dimension], check_path_changes=True
) -> Dimension:
    """Squash a list of nested Dimensions into a single one.

    Args:
        dimensions: The Dimensions to squash, from slowest to fastest moving
        check_path_changes: If True then check that nesting the output
            Dimension within other Dimensions will provide the same path
            as nesting the input Dimension within other Dimensions

    See Also:
        `why-squash-can-change-path`

    >>> dimx = Dimension({"x": np.array([1, 2])}, snake=True)
    >>> dimy = Dimension({"y": np.array([3, 4])})
    >>> squash_dimensions([dimy, dimx]).positions
    {'y': array([3, 3, 4, 4]), 'x': array([1, 2, 2, 1])}
    """
    path = Path(dimensions)
    # Comsuming a Path of these dimensions performs the squash
    # TODO: dim.tile might give better performance but is much longer
    squashed = path.consume()
    # Check that the squash is the same as the original
    if dimensions and dimensions[0].snake:
        squashed.snake = True
        # The top level is snaking, so this dimension will run backwards
        # This means any non-snaking axes will run backwards, which is
        # surprising, so don't allow it
        if check_path_changes:
            non_snaking = [k for d in dimensions for k in d.keys() if not d.snake]
            if non_snaking:
                raise ValueError(
                    f"Cannot squash non-snaking Specs in a snaking Dimension "
                    f"otherwise {non_snaking} would run backwards"
                )
    elif check_path_changes:
        # The top level is not snaking, so make sure there is an even
        # number of iterations of any snaking axis within it so it
        # doesn't jump when this dimension is iterated a second time
        for i, dim in enumerate(dimensions):
            # A snaking dimension within a non-snaking top level must repeat
            # an even number of times
            if dim.snake and np.product(path._lengths[:i]) % 2:
                raise ValueError(
                    f"Cannot squash snaking Specs in a non-snaking Dimension "
                    f"when they do not repeat an even number of times "
                    f"otherwise {dim.keys()} would jump in position"
                )
    return squashed


class Path:
    """Create a consumable Path through a list of Dimensions representing a
    scan path.

    Args:
        dimensions: The Dimensions describing the scan, from slowest to fastest
            moving
        start: The index of where in the Path to start
        num: The number of scan points to produce after start. None means up to
            the end

    See Also:
        `iterate-a-spec`
    """

    def __init__(
        self, dimensions: List[Dimension], start: int = 0, num: int = None,
    ):
        #: The Dimensions describing the scan, from slowest to fastest moving
        self.dimensions = dimensions
        #: Index that is next to be consumed
        self.index = start
        self._lengths = np.array([len(dim) for dim in dimensions])
        #: Index of the end point, one more than the last index that will be
        #: produced
        self.end_index = np.product(self._lengths)
        if num is not None and start + num < self.end_index:
            self.end_index = start + num

    def consume(self, num: int = None) -> Dimension:
        """Consume at most num points from the Path and return as a Dimension

        >>> dimx = Dimension({"x": np.array([1, 2])}, snake=True)
        >>> dimy = Dimension({"y": np.array([3, 4])})
        >>> path = Path([dimy, dimx])
        >>> path.consume(3).positions
        {'y': array([3, 3, 4]), 'x': array([1, 2, 2])}
        >>> path.consume(3).positions
        {'y': array([4]), 'x': array([1])}
        >>> path.consume(3).positions
        {'y': array([], dtype=int64), 'x': array([], dtype=int64)}
        """
        if num is None:
            end_index = self.end_index
        else:
            end_index = min(self.index + num, self.end_index)
        indices = np.arange(self.index, end_index)
        self.index = end_index
        positions, lower, upper = {}, {}, {}
        if len(indices) > 0:
            self.index = indices[-1] + 1
        # Example numbers below from a 2x3x4 ZxYxX scan
        for i, dim in enumerate(self.dimensions):
            # Number of times each position will repeat: Z:12, Y:4, X:1
            repeats = np.product(self._lengths[i + 1 :])
            # How big is this dim: Z:2, Y:3, X:4
            dim_len = self._lengths[i]
            # Scan indices mapped to indices within dimension:
            # Z:000000000000111111111111
            # Y:000011112222000011112222
            # X:012301230123012301230123
            dim_indices = (indices // repeats) % dim_len
            if dim.snake:
                # Whether this point is running backwards:
                # Z:000000000000000000000000
                # Y:000000000000111111111111
                # X:000011110000111100001111
                backwards = (indices // (repeats * dim_len)) % 2
                # The scan indices mapped to snaking ones:
                # Z:000000000000111111111111
                # Y:000011112222222211110000
                # X:012332100123321001233210
                snake_indices = np.where(
                    backwards, dim_len - 1 - dim_indices, dim_indices
                )
                for key in dim.keys():
                    positions[key] = dim.positions[key][snake_indices]
                    # If going backwards, select from the opposite bound
                    lower[key] = np.where(
                        backwards,
                        dim.upper[key][snake_indices],
                        dim.lower[key][snake_indices],
                    )
                    upper[key] = np.where(
                        backwards,
                        dim.lower[key][snake_indices],
                        dim.upper[key][snake_indices],
                    )
            else:
                for key in dim.keys():
                    positions[key] = dim.positions[key][dim_indices]
                    lower[key] = dim.lower[key][dim_indices]
                    upper[key] = dim.upper[key][dim_indices]
        return Dimension(positions, lower, upper)

    def __len__(self) -> int:
        """Number of points left in a scan, reduces when `consume` is called"""
        return self.end_index - self.index


class SpecPositions:
    """Convenience iterable that produces the scan positions for each axis. For
    better performance, consume from a `Path` instead.

    Args:
        dimensions: The Dimensions describing the scan, from slowest to fastest
            moving

    See Also:
        `iterate-a-spec`

    >>> dimx = Dimension({"x": np.array([1, 2])}, snake=True)
    >>> dimy = Dimension({"y": np.array([3, 4])})
    >>> sp = SpecPositions([dimy, dimx])
    >>> for p in sp: print(p)
    {'y': 3, 'x': 1}
    {'y': 3, 'x': 2}
    {'y': 4, 'x': 2}
    {'y': 4, 'x': 1}
    """

    def __init__(self, dimensions: List[Dimension]):
        #: The Dimensions describing the scan, from slowest to fastest moving
        self.dimensions = dimensions

    @property
    def keys(self) -> List:
        """The keys that will be present in each position dictionary"""
        keys = []
        for dim in self.dimensions:
            keys += dim.keys()
        return keys

    def __len__(self) -> int:
        """The number of dictionaries that will be produced if iterated over"""
        return np.product([len(dim) for dim in self.dimensions])

    def __iter__(self) -> Iterator[Positions]:
        path = Path(self.dimensions)
        while len(path):
            dim = path.consume(1)
            yield {k: dim.positions[k][0] for k in dim.keys()}



from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
from pydantic import Field, parse_obj_as, parse_raw_as, validate_arguments

from .core import (
    Dimension,
    Path,
    SpecPositions,
    _WithType,
    if_instance_do,
    squash_dimensions,
)
from .regions import Region, get_mask


class Spec(_WithType):
    """Abstract baseclass for the specification of a scan. Supports operators:

    - ``*``: Outer `Product` of two Specs, nesting the second within the first
    - ``+``: `Zip` two Specs together, iterating in tandem
    - ``&``: `Mask` the Spec with a `Region`, excluding positions outside it
    - ``~``: `Snake` the Spec, reversing every other iteration of it
    """

    def keys(self) -> List:
        """Implemented by subclasses to produce the list of keys that are
        present in the positions, from slowest moving to fastest moving"""
        raise NotImplementedError(self)

    def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
        """Implemented by subclasses to produce the `Dimension` list that
        contribute to positions, from slowest moving to fastest moving"""
        raise NotImplementedError(self)

    def path(self) -> Path:
        """Return a `Path` through the scan that can be consumed in chunks
        to give positions and bounds"""
        path = Path(self.create_dimensions())
        return path

    def positions(self) -> SpecPositions:
        """Return a `SpecPositions` that can be iterated position by position"""
        sp = SpecPositions(self.create_dimensions(bounds=False))
        return sp

    def __mul__(self, other) -> "Product":
        return if_instance_do(other, Spec, lambda o: Product(self, o))

    def __add__(self, other) -> "Zip":
        return if_instance_do(other, Spec, lambda o: Zip(self, o))

    def __and__(self, other) -> "Mask":
        return if_instance_do(other, Region, lambda o: Mask(self, o))

    def __invert__(self) -> "Snake":
        return Snake(self)


class Product(Spec):
    """Outer product of two Specs, nesting inner within outer. This means that
    inner will run in its entirety at each point in outer.

    .. example_spec::

        from scanspec.specs import Line

        spec = Line("y", 1, 2, 3) * Line("x", 3, 4, 12)
    """

    outer: Spec = Field(..., description="Will be executed once")
    inner: Spec = Field(..., description="Will be executed len(outer) times")

    def keys(self) -> List:
        return self.outer.keys() + self.inner.keys()

    def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
        dims_outer = self.outer.create_dimensions(bounds=False, nested=nested)
        dims_inner = self.inner.create_dimensions(bounds, nested=True)
        return dims_outer + dims_inner


class Zip(Spec):
    """Run two Specs in parallel, merging their positions together. Typically
    formed using the ``+`` operator.

    Dimensions are merged by:

    - If right creates a single Dimension of size 1, expand it to the size of
      the fastest Dimension created by left
    - Merge individual dimensions together from fastest to slowest

    This means that Zipping a Spec producing Dimensions [l2, l1] with a
    Spec producing Dimension [r1] will assert len(l1)==len(r1), and produce
    Dimensions [l2, l1+r1].

    .. example_spec::

        from scanspec.specs import Line

        spec = Line("z", 1, 2, 3) * Line("y", 3, 4, 5) + Line("x", 4, 5, 5)
    """

    left: Spec = Field(
        ..., description="The left-hand Spec to Zip, will appear earlier in keys"
    )
    right: Spec = Field(
        ..., description="The right-hand Spec to Zip, will appear later in keys"
    )

    def keys(self) -> List:
        return self.left.keys() + self.right.keys()

    def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
        dims_left = self.left.create_dimensions(bounds, nested)
        dims_right = self.right.create_dimensions(bounds, nested)
        assert len(dims_left) >= len(
            dims_right
        ), f"Zip requires len({self.left}) >= len({self.right})"

        # Pad and expand the right to be the same size as left
        # Special case, if only one dim with size 1, expand to the right size
        if len(dims_right) == 1 and len(dims_right[0]) == 1:
            repeated = dims_right[0].repeat(len(dims_left[-1]))
            repeated.snake = dims_left[-1].snake
            dims_right = [repeated]

        # Left pad dims_right with Nones so they are the same size
        npad = len(dims_left) - len(dims_right)
        padded_right: List[Optional[Dimension]] = [None] * npad
        padded_right += dims_right

        # Work through, zipping them together one by one
        dimensions = []
        for dim_left, dim_right in zip(dims_left, padded_right):
            if dim_right is None:
                dim = dim_left
            else:
                dim = dim_left.zip(dim_right)
            assert isinstance(
                dim, Dimension
            ), f"Padding went wrong {dims_left} {padded_right}"
            dimensions.append(dim)
        return dimensions


class Mask(Spec):
    """Restrict the given Spec to only the positions that fall inside the given
    Region.

    Typically created with the ``&`` operator. It also pushes down the
    ``& | ^ -`` operators to its `Region` to avoid the need for brackets on
    combinations of Regions.

    If a Region spans multiple Dimensions, these Dimensions will be squashed
    together.

    See Also:
        `why-squash-can-change-path`

    .. example_spec::

        from scanspec.specs import Line
        from scanspec.regions import Circle

        spec = Line("y", 1, 3, 3) * Line("x", 3, 5, 5) & Circle("x", "y", 4, 2, 1.2)
    """

    spec: Spec = Field(..., description="The Spec containing the source positions")
    region: Region = Field(..., description="The Region that positions will be inside")
    check_path_changes: bool = Field(
        True, description="If True path through scan will not be modified by squash"
    )

    def keys(self) -> List:
        return self.spec.keys()

    def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
        dims = self.spec.create_dimensions(bounds, nested)
        for key_set in self.region.key_sets():
            # Squash the dimensions together containing these keys
            matches = [i for i, d in enumerate(dims) if set(d.keys()) & key_set]
            assert matches, f"No Specs match keys {list(key_set)}"
            si, ei = matches[0], matches[-1]
            if si != ei:
                # Span Specs, squash them
                # If the spec to be squashed is nested (inside the Mask or outside)
                # then check the path changes if requested
                check_path_changes = (nested or si) and self.check_path_changes
                squashed = squash_dimensions(dims[si : ei + 1], check_path_changes)
                dims = dims[:si] + [squashed] + dims[ei + 1 :]
        # Generate masks from the positions showing what's inside
        masked_dims = [dim.mask(get_mask(self.region, dim.positions)) for dim in dims]
        return masked_dims

    # *+ bind more tightly than &|^ so without these overrides we
    # would need to add brackets to all combinations of Regions
    def __or__(self, other: "Region") -> "Mask":
        return if_instance_do(other, Region, lambda o: Mask(self.spec, self.region | o))

    def __and__(self, other: "Region") -> "Mask":
        return if_instance_do(other, Region, lambda o: Mask(self.spec, self.region & o))

    def __xor__(self, other: "Region") -> "Mask":
        return if_instance_do(other, Region, lambda o: Mask(self.spec, self.region ^ o))

    # This is here for completeness, tends not to be called as - binds
    # tighter than &
    def __sub__(self, other: "Region") -> "Mask":
        return if_instance_do(other, Region, lambda o: Mask(self.spec, self.region - o))


class Snake(Spec):
    """Run the Spec in reverse on every other iteration when nested inside
    another Spec. Typically created with the ``~`` operator.

    .. example_spec::

        from scanspec.specs import Line

        spec = Line("y", 1, 3, 3) * ~Line("x", 3, 5, 5)
    """

    spec: Spec = Field(
        ..., description="The Spec to run in reverse every other iteration"
    )

    def keys(self) -> List:
        return self.spec.keys()

    def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
        dims = self.spec.create_dimensions(bounds, nested)
        for dim in dims:
            dim.snake = True
        return dims


class Concat(Spec):
    """Concatenate two Specs together, running one after the other. Each Dimension
    of left and right must contain the same keys.

    .. example_spec::

        from scanspec.specs import Line, Concat

        spec = Concat(Line("x", 1, 3, 3), Line("x", 4, 5, 5))
    """

    left: Spec = Field(
        ..., description="The left-hand Spec to Zip, positions will appear earlier"
    )
    right: Spec = Field(
        ..., description="The right-hand Spec to Zip, positions will appear later"
    )

    def keys(self) -> List:
        left_keys, right_keys = self.left.keys(), self.right.keys()
        assert left_keys == right_keys, f"Keys {left_keys} != {right_keys}"
        return left_keys

    def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
        dims_left = self.left.create_dimensions(bounds, nested)
        dims_right = self.right.create_dimensions(bounds, nested)
        assert len(dims_right) == len(
            dims_left
        ), f"Specs {self.left} and {self.right} don't have same number of dimensions"
        dimensions = [dl.concat(dr) for dl, dr in zip(dims_left, dims_right)]
        return dimensions


class Squash(Spec):
    """Squash the Dimensions together of the scan (but not positions) into one
    long line.

    See Also:
        `why-squash-can-change-path`

    .. example_spec::

        from scanspec.specs import Line, Squash

        spec = Squash(Line("y", 1, 2, 3) * Line("x", 0, 1, 4))
    """

    spec: Spec = Field(..., description="The Spec to squash the dimensions of")
    check_path_changes: bool = Field(
        True, description="If True path through scan will not be modified by squash"
    )

    def keys(self) -> List:
        return self.spec.keys()

    def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
        # TODO: if we squash we explode the size, can we avoid this?
        dims = self.spec.create_dimensions(bounds, nested)
        dim = squash_dimensions(dims, nested and self.check_path_changes)
        return [dim]


def _dimensions_from_indexes(
    func: Callable[[np.ndarray], Dict[Any, np.ndarray]],
    keys: List,
    num: int,
    bounds: bool,
) -> List[Dimension]:
    positions_calc = func(np.linspace(0.5, num - 0.5, num))
    positions = {k: positions_calc[k] for k in keys}
    if bounds:
        bounds_calc = func(np.linspace(0, num, num + 1))
        lower = {k: bounds_calc[k][:-1] for k in keys}
        upper = {k: bounds_calc[k][1:] for k in keys}
        dimension = Dimension(positions, lower, upper)
    else:
        dimension = Dimension(positions)
    return [dimension]


class Line(Spec):
    """Linearly spaced points in the given key, with first and last points
    centred on start and stop.

    .. example_spec::

        from scanspec.specs import Line

        spec = Line("x", 1, 2, 5)
    """

    # TODO: are start and stop positions, bounds, or different for fly/step
    key: Any = Field(..., description="An identifier for what to move")
    start: float = Field(..., description="Centre point of the first point of the line")
    stop: float = Field(..., description="Centre point of the last point of the line")
    num: int = Field(..., ge=1, description="Number of points to produce")

    def keys(self) -> List:
        return [self.key]

    def _line_from_indexes(self, indexes: np.ndarray) -> Dict[Any, np.ndarray]:
        if self.num == 1:
            # Only one point, stop-start gives length of one point
            step = self.stop - self.start
        else:
            # Multiple points, stop-start gives length of num-1 points
            step = (self.stop - self.start) / (self.num - 1)
        return {self.key: (indexes - 0.5) * step + self.start}

    def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
        return _dimensions_from_indexes(
            self._line_from_indexes, self.keys(), self.num, bounds
        )

    @classmethod
    @validate_arguments
    def bounded(
        cls,
        key=key,
        lower: float = Field(
            ..., description="Lower bound of the first point of the line"
        ),
        upper: float = Field(
            ..., description="Upper bound of the last point of the line"
        ),
        num: int = num,
    ):
        """Specify a Line by extreme bounds instead of centre points.

        .. example_spec::

            from scanspec.specs import Line

            spec = Line.bounded("x", 1, 2, 5)
        """
        half_step = (upper - lower) / num / 2
        start = lower + half_step
        if num == 1:
            # One point, stop will only be used for step size
            stop = upper + half_step
        else:
            # Many points, stop will be produced
            stop = upper - half_step
        return cls(key, start, stop, num)


class Static(Spec):
    """A static point, repeated "num" times, with "key" at "value". Can
    be used to set key=value at every point in a scan.

    .. example_spec::

        from scanspec.specs import Line, Static

        spec = Line("y", 1, 2, 3) + Static("x", 3)
    """

    key: Any = Field(..., description="An identifier for what to move")
    value: float = Field(..., description="The value at each point")
    num: int = Field(1, ge=1, description="How many times to repeat this point")

    def keys(self) -> List:
        return [self.key]

    def _repeats_from_indexes(self, indexes: np.ndarray) -> Dict[Any, np.ndarray]:
        return {self.key: np.full(len(indexes), self.value)}

    def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
        return _dimensions_from_indexes(
            self._repeats_from_indexes, self.keys(), self.num, bounds
        )


class Spiral(Spec):
    """Archimedean spiral of "x_key" and "y_key", starting at centre point
    ("x_start", "y_start") with angle "rotate". Produces "num" points
    in a spiral spanning width of "x_range" and height of "y_range"

    .. example_spec::

        from scanspec.specs import Spiral

        spec = Spiral("x", "y", 1, 5, 10, 50, 30)
    """

    x_key: Any = Field(..., description="An identifier for what to move for x")
    y_key: Any = Field(..., description="An identifier for what to move for y")
    # TODO: do we like these names?
    x_start: float = Field(..., description="x centre of the spiral")
    y_start: float = Field(..., description="y centre of the spiral")
    x_range: float = Field(..., description="x width of the spiral")
    y_range: float = Field(..., description="y width of the spiral")
    num: int = Field(..., description="Number of points in the spiral")
    rotate: float = Field(0.0, description="How much to rotate the angle of the spiral")

    def keys(self) -> List:
        # TODO: reversed from __init__ args, a good idea?
        return [self.y_key, self.x_key]

    def _spiral_from_indexes(self, indexes: np.ndarray) -> Dict[Any, np.ndarray]:
        # simplest spiral equation: r = phi
        # we want point spacing across area to be the same as between rings
        # so: sqrt(area / num) = ring_spacing
        # so: sqrt(pi * phi^2 / num) = 2 * pi
        # so: phi = sqrt(4 * pi * num)
        phi = np.sqrt(4 * np.pi * indexes)
        # indexes are 0..num inclusive, and diameter is 2x biggest phi
        diameter = 2 * np.sqrt(4 * np.pi * self.num)
        # scale so that the spiral is strictly smaller than the range
        x_scale = self.x_range / diameter
        y_scale = self.y_range / diameter
        return {
            self.y_key: self.y_start + y_scale * phi * np.cos(phi + self.rotate),
            self.x_key: self.x_start + x_scale * phi * np.sin(phi + self.rotate),
        }

    def create_dimensions(self, bounds=True, nested=False) -> List[Dimension]:
        return _dimensions_from_indexes(
            self._spiral_from_indexes, self.keys(), self.num, bounds
        )

    @classmethod
    @validate_arguments
    def spaced(
        cls,
        x_key: Any = x_key,
        y_key: Any = y_key,
        x_start: float = x_start,
        y_start: float = y_start,
        radius: float = Field(..., description="radius of the spiral"),
        dr: float = Field(..., description="difference between each ring"),
        rotate: float = rotate,
    ):
        """Specify a Spiral equally spaced in "x_key" and "y_key" by specifying
        the "radius" and difference between each ring of the spiral "dr"

        .. example_spec::

            from scanspec.specs import Spiral

            spec = Spiral.spaced("x", "y", 0, 0, 10, 3)
        """
        # phi = sqrt(4 * pi * num)
        # and: n_rings = phi / (2 * pi)
        # so: n_rings * 2 * pi = sqrt(4 * pi * num)
        # so: num = n_rings^2 * pi
        n_rings = radius / dr
        num = n_rings ** 2 * np.pi
        return cls(x_key, y_key, x_start, y_start, radius * 2, radius * 2, num, rotate)


#: Can be used as a special key to indicate how long each point should be
TIME = "TIME"


def fly(spec: Spec, duration: float):
    """Flyscan, zipping TIME=duration for every position

    .. example_spec::

        from scanspec.specs import Line, fly

        spec = fly(Line("x", 1, 2, 3), 0.1)
    """
    return spec + Static(TIME, duration)


def step(spec: Spec, duration: float):
    """Step scan, adding TIME=duration as an inner dimension for every position

    .. example_spec::

        from scanspec.specs import Line, step

        spec = step(Line("x", 1, 2, 3), 0.1)
    """
    return spec * Static(TIME, duration)


class _UnionModifier:
    # Modifies all Spec subclasses so Spec->Union[all Spec subclasses]
    def __init__(self):
        _spec_subclasses = tuple(self._all_subclasses(Spec))
        _region_subclasses = tuple(self._all_subclasses(Region))
        self.spec_union = Union[_spec_subclasses]  # type: ignore
        self.region_union = Union[_region_subclasses]  # type: ignore

        for spec in _spec_subclasses + _region_subclasses:
            for _, field in spec.__fields__.items():
                if field.type_ is Spec:
                    # TODO: Is it better to create a new field than modify?
                    field.type_ = self.spec_union
                    field.prepare()
                elif field.type_ is Region:
                    field.type_ = self.region_union
                    field.prepare()

    def _all_subclasses(self, cls):
        return set(cls.__subclasses__()).union(
            [s for c in cls.__subclasses__() for s in self._all_subclasses(c)]
        )


_modifier = _UnionModifier()


def spec_from_dict(d: Dict) -> Spec:
    return parse_obj_as(_modifier.spec_union, d)  # type: ignore


def spec_from_json(text: str) -> Spec:
    return parse_raw_as(_modifier.spec_union, text)  # type: ignore



from typing import Any, Dict, Iterator, List, Set

import numpy as np
from pydantic import Field
from pydantic.main import BaseModel

from .core import Positions, _WithType, if_instance_do


class Region(_WithType):
    """Abstract baseclass for a Region that can `Mask` a `Spec`. Supports operators:

    - ``|``: `UnionOf` two Regions, positions present in either
    - ``&``: `IntersectionOf` two Regions, positions present in both
    - ``-``: `DifferenceOf` two Regions, positions present in first not second
    - ``^``: `SymmetricDifferenceOf` two Regions, positions present in one not both
    """

    def key_sets(self) -> List[Set[str]]:
        """Implemented by subclasses to produce the non-overlapping sets of keys
        this region spans"""
        raise NotImplementedError(self)

    def mask(self, positions: Dict[Any, np.ndarray]) -> np.ndarray:
        """Implemented by subclasses to produce a mask of which positions are in
        the region"""
        raise NotImplementedError(self)

    def __or__(self, other) -> "UnionOf":
        return if_instance_do(other, Region, lambda o: UnionOf(self, o))

    def __and__(self, other) -> "IntersectionOf":
        return if_instance_do(other, Region, lambda o: IntersectionOf(self, o))

    def __sub__(self, other) -> "DifferenceOf":
        return if_instance_do(other, Region, lambda o: DifferenceOf(self, o))

    def __xor__(self, other) -> "SymmetricDifferenceOf":
        return if_instance_do(other, Region, lambda o: SymmetricDifferenceOf(self, o))


def get_mask(region: Region, positions: Positions) -> np.ndarray:
    """If there is an overlap of keys of region and positions return a
    mask of the positions in the region, otherwise return all ones
    """
    keys = set(positions)
    needs_mask = any(ks & keys for ks in region.key_sets())
    if needs_mask:
        return region.mask(positions)
    else:
        return np.ones(len(list(positions.values())[0]))


def _merge_key_sets(key_sets: List[Set[str]]) -> Iterator[Set[str]]:
    # Take overlapping key sets and merge any that overlap into each
    # other
    for ks in key_sets:
        key_set = ks.copy()
        # Empty matching sets into this key_set
        for ks in key_sets:
            if ks & key_set:
                while ks:
                    key_set.add(ks.pop())
        # It might be emptied already, only yield if it isn't
        if key_set:
            yield key_set


class CombinationOf(Region):
    """Abstract baseclass for a combination of two regions, left and right"""

    left: Region = Field(..., description="The left-hand Region to combine")
    right: Region = Field(..., description="The right-hand Region to combine")

    def key_sets(self) -> List[Set[str]]:
        key_sets = list(_merge_key_sets(self.left.key_sets() + self.right.key_sets()))
        return key_sets


# Naming so we don't clash with typing.Union
class UnionOf(CombinationOf):
    """A position is in UnionOf(a, b) if it is in either a or b. Typically
    created with the ``|`` operator

    >>> r = Range("x", 0.5, 2.5) | Range("x", 1.5, 3.5)
    >>> r.mask({"x": np.array([0, 1, 2, 3, 4])})
    array([False,  True,  True,  True, False])
    """

    def mask(self, positions: Dict[Any, np.ndarray]) -> np.ndarray:
        mask = get_mask(self.left, positions) | get_mask(self.right, positions)
        return mask


class IntersectionOf(CombinationOf):
    """A position is in IntersectionOf(a, b) if it is in both a and b. Typically
    created with the ``&`` operator

    >>> r = Range("x", 0.5, 2.5) & Range("x", 1.5, 3.5)
    >>> r.mask({"x": np.array([0, 1, 2, 3, 4])})
    array([False, False,  True, False, False])
    """

    def mask(self, positions: Dict[Any, np.ndarray]) -> np.ndarray:
        mask = get_mask(self.left, positions) & get_mask(self.right, positions)
        return mask


class DifferenceOf(CombinationOf):
    """A position is in DifferenceOf(a, b) if it is in a and not in b. Typically
    created with the ``-`` operator

    >>> r = Range("x", 0.5, 2.5) - Range("x", 1.5, 3.5)
    >>> r.mask({"x": np.array([0, 1, 2, 3, 4])})
    array([False,  True, False, False, False])
    """

    def mask(self, positions: Dict[Any, np.ndarray]) -> np.ndarray:
        left_mask = get_mask(self.left, positions)
        # Return the xor restricted to the left region
        mask = left_mask ^ get_mask(self.right, positions) & left_mask
        return mask


class SymmetricDifferenceOf(CombinationOf):
    """A position is in SymmetricDifferenceOf(a, b) if it is in either a or b,
    but not both. Typically created with the ``^`` operator

    >>> r = Range("x", 0.5, 2.5) ^ Range("x", 1.5, 3.5)
    >>> r.mask({"x": np.array([0, 1, 2, 3, 4])})
    array([False,  True, False,  True, False])
    """

    def mask(self, positions: Dict[Any, np.ndarray]) -> np.ndarray:
        mask = get_mask(self.left, positions) ^ get_mask(self.right, positions)
        return mask


class Range(Region):
    """Mask contains positions of key >= min and <= max

    >>> r = Range("x", 1, 2)
    >>> r.mask({"x": np.array([0, 1, 2, 3, 4])})
    array([False,  True,  True, False, False])
    """

    key: Any = Field(..., description="The key matching the axis to mask in the spec")
    min: float = Field(..., description="The minimum inclusive value in the region")
    max: float = Field(..., description="The minimum inclusive value in the region")

    def key_sets(self) -> List[Set[str]]:
        return [{self.key}]

    def mask(self, positions: Dict[Any, np.ndarray]) -> np.ndarray:
        v = positions[self.key]
        mask = np.bitwise_and(v >= self.min, v <= self.max)
        return mask


class Rectangle(Region):
    """Mask contains positions of key within a rotated xy rectangle

    .. example_spec::

        from scanspec.specs import Line
        from scanspec.regions import Rectangle

        grid = Line("y", 1, 3, 10) * ~Line("x", 0, 2, 10)
        spec = grid & Rectangle("x", "y", 0, 1.1, 1.5, 2.1, 30)
    """

    x_key: Any = Field(..., description="The key matching the x axis of the spec")
    y_key: Any = Field(..., description="The key matching the y axis of the spec")
    x_min: float = Field(..., description="Minimum inclusive x value in the region")
    y_min: float = Field(..., description="Minimum inclusive y value in the region")
    x_max: float = Field(..., description="Maximum inclusive x value in the region")
    y_max: float = Field(..., description="Maximum inclusive y value in the region")
    angle: float = Field(0, description="Clockwise rotation angle of the rectangle")

    def key_sets(self) -> List[Set[str]]:
        return [{self.x_key, self.y_key}]

    def mask(self, positions: Dict[Any, np.ndarray]) -> np.ndarray:
        x = positions[self.x_key] - self.x_min
        y = positions[self.y_key] - self.y_min
        if self.angle != 0:
            # Rotate src positions by -angle
            phi = np.radians(-self.angle)
            rx = x * np.cos(phi) - y * np.sin(phi)
            ry = x * np.sin(phi) + y * np.cos(phi)
            x = rx
            y = ry
        mask_x = np.bitwise_and(x >= 0, x <= (self.x_max - self.x_min))
        mask_y = np.bitwise_and(y >= 0, y <= (self.y_max - self.y_min))
        return mask_x & mask_y


class Circle(Region):
    """Mask contains positions of key within an xy circle of given radius

    .. example_spec::

        from scanspec.specs import Line
        from scanspec.regions import Circle

        grid = Line("y", 1, 3, 10) * ~Line("x", 0, 2, 10)
        spec = grid & Circle("x", "y", 1, 2, 0.9)
    """

    x_key: Any = Field(..., description="The key matching the x axis of the spec")
    y_key: Any = Field(..., description="The key matching the x axis of the spec")
    x_centre: float = Field(
        ..., description="Minimum inclusive x value in the region", alias="x_center"
    )
    y_centre: float = Field(
        ..., description="Minimum inclusive y value in the region", alias="y_center"
    )
    radius: float = Field(..., description="Radius of the circle")

    class Config:
        # Allow either centre or center
        allow_population_by_field_name = True

    def key_sets(self) -> List[Set[str]]:
        return [{self.x_key, self.y_key}]

    def mask(self, positions: Dict[Any, np.ndarray]) -> np.ndarray:
        x = positions[self.x_key] - self.x_centre
        y = positions[self.y_key] - self.y_centre
        mask = x * x + y * y <= (self.radius * self.radius)
        return mask


def find_regions(obj) -> Iterator[Region]:
    """Recursively iterate over obj and its children, yielding any Region
    instances found"""
    if isinstance(obj, BaseModel):
        if isinstance(obj, Region):
            yield obj
        for name in obj.__fields__:
            yield from find_regions(getattr(obj, name))



from itertools import cycle
from typing import Any, Dict, List

import numpy as np
from matplotlib import colors, patches
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import proj3d
from scipy import interpolate

from .core import Dimension, Path
from .regions import Circle, Rectangle, find_regions
from .specs import TIME, Spec


def _find_breaks(dim: Dimension):
    breaks = []
    for key in dim.keys():
        breaks.append(dim.lower[key][1:] != dim.upper[key][:-1])
    same = np.logical_or.reduce(breaks)
    break_indices = np.nonzero(same)[0] + 1
    return list(break_indices)


def _plot_arrays(axes, arrays: List[np.ndarray], **kwargs):
    if len(arrays) > 2:
        axes.plot3D(arrays[2], arrays[1], arrays[0], **kwargs)
    elif len(arrays) == 2:
        axes.plot(arrays[1], arrays[0], **kwargs)
    else:
        axes.plot(arrays[0], np.zeros(len(arrays[0])), **kwargs)


# https://stackoverflow.com/a/11156353
class _Arrow3D(patches.FancyArrowPatch):
    def __init__(self, xs, ys, zs, *args, **kwargs):
        super().__init__((0, 0), (0, 0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def draw(self, renderer):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, _ = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        super().draw(renderer)


def _plot_arrow(axes, arrays: List[np.ndarray]):
    if len(arrays) == 1:
        arrays = [[0, 0]] + arrays
    if len(arrays) == 2:
        head = [a[-1] for a in reversed(arrays)]
        tail = [a[-1] - (a[-1] - a[-2]) * 0.1 for a in reversed(arrays)]
        axes.annotate(
            "", head[:2], tail[:2], arrowprops=dict(color="lightgrey", arrowstyle="-|>")
        )
    elif len(arrays) == 3:
        arrows = [a[-2:] for a in reversed(arrays)]
        a = _Arrow3D(
            *arrows[:3], mutation_scale=10, arrowstyle="-|>", color="lightgrey"
        )
        axes.add_artist(a)


def _plot_spline(axes, ranges, arrays: List[np.ndarray], index_colours: Dict[int, str]):
    scaled_arrays = [a / r for a, r in zip(arrays, ranges)]
    # Define curves parametrically
    t = np.zeros(len(arrays[0]))
    t[1:] = np.sqrt(sum((arr[1:] - arr[:-1]) ** 2 for arr in scaled_arrays))
    t = np.cumsum(t)
    if t[-1] > 0:
        # There are no duplicated points, plot a spline
        t /= t[-1]
        # Scale the arrays so splines don't favour larger scaled axes
        tck, _ = interpolate.splprep(scaled_arrays, k=2, s=0)
        starts = sorted(list(index_colours))
        stops = starts[1:] + [len(arrays[0]) - 1]
        for start, stop in zip(starts, stops):
            tnew = np.linspace(t[start], t[stop], num=1001)
            spline = interpolate.splev(tnew, tck)
            # Scale the splines back to the original scaling
            unscaled_splines = [a * r for a, r in zip(spline, ranges)]
            _plot_arrays(axes, unscaled_splines, color=index_colours[start])
            yield unscaled_splines


def plot_spec(spec: Spec):
    """Plot a spec, drawing the path taken through the scan, using a different
    colour for each point, grey for the turnarounds, and marking the
    centrepoints with a filled circle if there are less than 200 of them. If the
    scan is 2D then 2D regions are shown in black.

    .. example_spec::

        from scanspec.specs import Line
        from scanspec.regions import Circle

        cube = Line("z", 1, 3, 3) * Line("y", 1, 3, 10) * ~Line("x", 0, 2, 10)
        spec = cube & Circle("x", "y", 1, 2, 0.9)
    """
    dims = spec.create_dimensions()
    dim = Path(dims).consume()
    keys = [k for k in spec.keys() if k is not TIME]
    ndims = len(keys)

    # Setup axes
    if ndims > 2:
        plt.figure(figsize=(6, 6))
        axes = plt.axes(projection="3d")
        axes.grid(False)
        axes.set_zlabel(keys[-3])
        axes.set_ylabel(keys[-2])
        axes.view_init(elev=15)
    elif ndims == 2:
        plt.figure(figsize=(6, 6))
        axes = plt.axes()
        axes.set_ylabel(keys[-2])
    else:
        plt.figure(figsize=(6, 2))
        axes = plt.axes()
        axes.yaxis.set_visible(False)
    axes.set_xlabel(keys[-1])

    # Title with dimension sizes
    plt.title(", ".join(f"Dim[{' '.join(d.keys())} len={len(d)}]" for d in dims))

    # Plot any Regions
    if ndims <= 2:
        for region in find_regions(spec):
            if isinstance(region, Rectangle):
                xy = (region.x_min, region.y_min)
                width = region.x_max - region.x_min
                height = region.y_max - region.y_min
                axes.add_patch(
                    patches.Rectangle(xy, width, height, region.angle, fill=False)
                )
            elif isinstance(region, Circle):
                xy = (region.x_centre, region.y_centre)
                axes.add_patch(patches.Circle(xy, region.radius, fill=False))

    # Plot the splines
    tail: Any = {k: None for k in keys}
    ranges = [max(np.max(v) - np.min(v), 0.0001) for k, v in dim.positions.items()]
    seg_col = cycle(colors.TABLEAU_COLORS)
    last_index = 0
    splines = None
    for index in _find_breaks(dim) + [len(dim)]:
        num_points = index - last_index
        arrays = []
        turnaround = []
        for k in keys:
            # Add the lower and positions
            arr = np.empty(num_points * 2 + 1)
            arr[:-1:2] = dim.lower[k][last_index:index]
            arr[1::2] = dim.positions[k][last_index:index]
            arr[-1] = dim.upper[k][index - 1]
            arrays.append(arr)
            # Add the turnaround
            if tail[k] is not None:
                # Already had a tail, add lead in points
                tail[k][2:] = np.linspace(-0.01, 0, 2) * (arr[1] - arr[0]) + arr[0]
                turnaround.append(tail[k])
            # Add tail off points
            tail[k] = np.empty(4)
            tail[k][:2] = np.linspace(0, 0.01, 2) * (arr[-1] - arr[-2]) + arr[-1]
        last_index = index

        arrow_arr = None
        if turnaround:
            # If we didn't move then plot a straight line from start to stop
            if all(t[1] - t[0] == 0 for t in turnaround):
                for t in turnaround:
                    t[1] += (t[2] - t[1]) / 4
            if all(t[3] - t[2] == 0 for t in turnaround):
                for t in turnaround:
                    t[2] -= (t[2] - t[1]) / 4
            # Plot the turnaround
            arrow_arr = list(_plot_spline(axes, ranges, turnaround, {0: "lightgrey"}))[
                0
            ]

        # Plot the points
        index_colours = {2 * i: next(seg_col) for i in range(num_points)}
        splines = list(_plot_spline(axes, ranges, arrays, index_colours))

        if arrow_arr:
            # Plot the arrow on the turnaround
            _plot_arrow(axes, arrow_arr)
        elif splines:
            # Plot the starting arrow in the direction of the first point
            arrow_arr = [(2 * a[0] - a[1], a[0]) for a in splines[0]]
            _plot_arrow(axes, arrow_arr)
        else:
            # First point isn't moving, put a right caret marker
            _plot_arrays(
                axes, [[dim.lower[k][0]] for k in keys], marker=5, color="lightgrey"
            )

    # Plot the capture points
    if len(dim) < 200:
        arrays = [dim.positions[k] for k in keys]
        _plot_arrays(axes, arrays, linestyle="", marker=".", color="k")

    # Plot the end
    _plot_arrays(
        axes, [[dim.upper[k][-1]] for k in keys], marker="x", color="lightgrey"
    )

    plt.show()
